#! /bin/sh
# mpath.sh Version 1.1 - 24-MAY-2004
#
# Manage setups with connections to multiple ISPs
# Julian Anastasov - May 2004

base="`dirname "$0"`"
[ "$base" != "" ] && base="`cd "$base" && pwd`"
[ "$base" = "" ] && exit 1

####################### Functions ############################

# Load configuration
# Arg1: configuration part: "main", "nh", etc
function load_config() {
	local f

	if [ "$1" = "main" ]
	then
		if ! [ -e "$base/mpath_cfg.src" ]
		then
			echo "Configuration file $base/mpath_cfg.src not found" > /dev/stderr
			exit 1
		fi
		source "$base/mpath_cfg.src"
		return 0
	fi
	if [ "$1" = "nh" ]
	then
		for f in $base/mpath_cfg_nh*.src
		do
			[ -e "$f" ] && source "$f"
		done
		return 0
	fi
	return 1
}

# Wrap the ip utility
function ip() {
	echo "$IP_TOOL $*"
	eval $IP_TOOL "$*"
}

# Lock file
# Arg1: filename to lock
# Arg2: timeout in seconds
function lock_file() {

	lockfile -1 -r 0 "$1" &> /dev/null
	[ "$?" -eq 0 ] && return 0
	echo "File $1 already locked, waiting $2 second(s)" > /dev/stderr
	[ "$2" -le 0 ] && return 1
	lockfile -1 -r "$2" -l 60 -s 20 "$1" &> /dev/null
	[ "$?" -eq 0 ] && return 0
	echo "Unable to lock file $1" > /dev/stderr
	return 1
}

# Do all needed NH checks and report for any changed NHs
function do_nh_check() {
	local n
	local nh
	local gw
	local ret=""

	n="${#NHDEV[@]}"
	for ((nh=0; nh<n; nh+=1))
	do
		gw="${NHGWPING[$nh]}"
		[ "$gw" = "" ] && continue
		ping -n -c 1 -w 1 -l 3 -I "${NHDEV[$nh]}" "$gw" &> /dev/null
		if [ "$?" -eq 0 ]
		then
			[ -e "$mpath_data/nh_down.$nh" ] || continue
			rm -f "$mpath_data/nh_down.$nh"
		else
			[ -e "$mpath_data/nh_down.$nh" ] && continue
			touch "$mpath_data/nh_down.$nh"
		fi
		if [ "$ret" = "" ]
		then
			ret="$nh"
		else
			ret="$ret $nh"
		fi
	done
	[ "$ret" != "" ] && echo "$ret"
	return 0
}

# Keep the ARP cache state for NHs valid
function do_nh_keeparp() {
	local n
	local nh
	local gw

	n="${#NHDEV[@]}"
	for ((nh=0; nh<n; nh+=1))
	do
		gw="${NHGWKEEP[$nh]}"
		[ "$gw" = "" ] && continue
		gw="${NHGW[$i]}"
		[ "$gw" = "" ] && continue
		ping -n -c 1 -w 1 -I "${NHDEV[$nh]}" "$gw" &> /dev/null
	done
}

# Apply settings after a Next Hop state change
# Arg1: NH ID
function do_nh_changed() {
	local nhs
	local nh
	local nh1
	local weight
	local j
	local i
	local i1
	local i2
	local n
	local n1
	local n2
	local rt1
	local rt2
	local src
	local fl
	local gw
	local best_nh
	local use_best_nh
	local nh_rule
	declare -a nharr
	declare -a src1

	if [ "$1" != "" ]
	then
		nhs="$1"
	else
		nhs="${#NHDEV[@]}"
		if [ "$nhs" -gt 0 ]
		then
			nhs="`seq 0 $[$nhs-1]`"
		else
			nhs=""
		fi
	fi
	[ "$nhs" = "" ] && return 1

	# Remove default route in case someone adds it
	[ "$test" = "" ] && ip route del default metric 0

	# Walk all routing tables and recreate all routes
	# containing this next hop

	n="${#TID[@]}"
	# Walk all tables
	for ((i=0; i<n; i+=1))
	do
		fl=0
		best_nh=""
		use_best_nh="y"
		for ((j=0; j<2; j+=1))
		do
			rt1=(${TROUTES[$i]})
			src1=(${TPREFSRC[$i]})
			n1="${#rt1[@]}"
			# Walk all routes in table
			for ((i1=0; i1<n1; i1+=1))
			do
				if [ "$j" -gt 0 ]
				then
					[ "${nharr[$i1]}" != "" ] && ip route append default table "${TID[$i]}" proto static "${nharr[$i1]}"
					continue
				fi
				nharr[$i1]=""
				src="${src1[$i1]}"
				rt2=(${rt1[$i1]//,/ })
				n2="${#rt2[@]}"
				# Walk all next hops in route
				for ((i2=0; i2<n2; i2+=1))
				do
					weight="${rt2[$i2]##*:}"
					if [ "$weight" = "${rt2[$i2]}" ]
					then
						weight=1
						nh="${rt2[$i2]}"
					else
						nh="${rt2[$i2]%%:*}"
					fi
					if [ "$fl" -eq 0 ]
					then
						for nh1 in $nhs
						do
							[ "$nh" != "$nh1" ] && continue
							fl=1
							break
						done
					fi

					# Check NH state:

					# weight=0 => NH is stopped for this route
					[ "$weight" -le 0 ] && continue

					# The NH's device supports IP protocol?
					[ -e "/proc/sys/net/ipv4/conf/${NHDEV[$nh]}" ] || continue

					gw="${NHGW[$nh]}"
					[ "$gw" != "" ] && gw="via $gw "
					nh_rule="nexthop ${gw}dev ${NHDEV[$nh]} weight $weight"

					# NH globally stopped?
					if [ -e "$mpath_data/nh_down.$nh" ]
					then
						# Remember the first failed next hop
						if [ "$best_nh" = "" -a "$use_best_nh" != "" ]
						then
							best_nh="$nh_rule"
							[ "$src" != "" -a "$src" != "." -a "$src" != "0" -a "$src" != "0.0.0.0" ] && best_nh="src $src $best_nh"
						fi
						continue
					fi

					# We have valid NH, do not use best_nh
					use_best_nh=""
					nharr[$i1]="${nharr[$i1]} $nh_rule"
				done
				# prefsrc for the route?
				[ "${nharr[$i1]}" != "" -a "$src" != "" -a "$src" != "." -a "$src" != "0" -a "$src" != "0.0.0.0" ] && nharr[$i1]="src $src ${nharr[$i1]}"
			done
			[ "$fl" -eq 0 -o "$j" -gt 0 ] && break
			for ((i1=0; i1<n1; i1+=1))
			do
				ip route del default metric 0 table "${TID[$i]}"
				# Optimization:
				[ "$?" -ne 0 ] && break
			done
			ip route add prohibit default proto static metric 1 table "${TID[$i]}" 2> /dev/null
		done
		# If all next hops are down then keep only one failed nexthop
		if [ "$fl" -ne 0 -a "$use_best_nh" != "" -a "$best_nh" != "" ]
		then
			ip route append default table "${TID[$i]}" proto static "$best_nh"
		fi
	done
	return 0
}

# Apply settings
# Arg1: command ("add"/"del")
function do_config() {
	local i
	local n
	local i1
	local n1
	local rt1
	local s
	local prio
	local gw
	local changed

	# IP Addresses
	n="${#AIP[@]}"
	for ((i=0; i<n; i+=1))
	do
		ip addr $1 ${AIP[$i]}
	done

	# IP Rules

	ip rule $1 prio 50 table main

	# Starting priority:
	prio=100

	n="${#IRULE[@]}"
	for ((i=0; i<n; i+=1))
	do
		ip rule $1 prio $prio ${IRULE[$i]}
		let prio+=1
	done

	if [ "$1" = "add" ]
	then
		# Mark all next hops DOWN before checking them

		n="${#NHDEV[@]}"
		for ((i=0; i<n; i+=1))
		do
			rm -f "$mpath_data/nh_down.$i"
			gw="${NHGWPING[$i]}"
			[ "$gw" = "" ] && continue
			touch "$mpath_data/nh_down.$i"
		done

		# Create routes in all tables
		# Activate at least one route per table to allow checks
		do_nh_changed ""

		ip route flush cache

		# Start with ARP probes
		do_nh_keeparp

		# Finally do some additional checks
		changed="`do_nh_check`"
		[ "$changed" != "" ] && do_nh_changed "$changed"

	else
		# Delete all tables

		n="${#TID[@]}"
		for ((i=0; i<n; i+=1))
		do
			rt1=(${TROUTES[$i]})
			n1="${#rt1[@]}"
			for ((i1=0; i1<n1; i1+=1))
			do
				ip route del default metric 0 table "${TID[$i]}"
				[ "$?" -ne 0 ] && break
			done
			ip route del default metric 1 table "${TID[$i]}"
		done
	fi
}

function list_settings() {

	echo "Defined settings:"

	n="${#AIP[@]}"
	nln=""
	[ "$n" -le 0 ] && nln=" N/A"
	echo " + IP Addresses:$nln"
	for ((i=0; i<n; i+=1))
	do
		echo "  - ${AIP[$i]}"
	done

	n="${#NHDEV[@]}"
	nln=""
	[ "$n" -le 0 ] && nln=" N/A"
	echo " + Next Hops:$nln"
	for ((i=0; i<n; i+=1))
	do
		gw="${NHGW[$i]}"
		[ "$gw" != "" ] && gw="GW $gw "
		keep="${NHGWKEEP[$i]}"
		[ "$keep" != "" ] && keep=" keep-arp"
		check="${NHGWPING[$i]}"
		[ "$check" != "" ] && check=" ping-check $check"
		arr_nh[$i]="${gw}dev ${NHDEV[$i]}$keep$check"
		echo "  - ${arr_nh[$i]}"
	done

	n="${#IRULE[@]}"
	nln=""
	[ "$n" -le 0 ] && nln=" N/A"
	echo " + IP Rules:$nln"
	for ((i=0; i<n; i+=1))
	do
		echo "  - ${IRULE[$i]}"
	done

	n="${#TID[@]}"
	nln=""
	[ "$n" -le 0 ] && nln=" N/A"
	echo " + Routing tables:$nln"
	for ((i=0; i<n; i+=1))
	do
		rt1=(${TROUTES[$i]})
		src1=(${TPREFSRC[$i]})
		n1="${#rt1[@]}"
		nln=""
		[ "$n1" -le 0 ] && nln=" empty"
		echo "  + table ${TID[$i]}:$nln"
		for ((i1=0; i1<n1; i1+=1))
		do
			rt2=(${rt1[$i1]//,/ })
			n2="${#rt2[@]}"
			if [ "$n2" -gt 1 ]
			then
				rttype="multipath"
			else
				rttype="unipath"
			fi
			src="${src1[$i1]}"
			[ "$src" != "" ] && src=", src $src"
			nln=""
			[ "$n2" -le 0 ] && nln=" empty"
			echo "   + $rttype route #$i1$src:$nln"
			for ((i2=0; i2<n2; i2+=1))
			do
				weight="${rt2[$i2]##*:}"
				if [ "$weight" = "${rt2[$i2]}" ]
				then
					if [ "$n2" -ge 2 ]
					then
						weight=", W=1"
					else
						weight=""
					fi
					nh="${rt2[$i2]}"
				else
					weight=", W=$weight"
					nh="${rt2[$i2]%%:*}"
				fi
				echo "    - NH #$nh$weight: ${arr_nh[$nh]}"
			done
		done
	done
}

####################### Main logic ############################

load_config main

if [ "$state_lock" = "" ]
then
	echo "Invalid configuration" > /dev/stderr
	exit 1
fi

if [ "$1" != "" -a "$1" != "list" ]
then
	slock=""
	if [ "$1" != "start" ]
	then
		if ! [ -e "$run_lock" ]
		then
			echo "$1: mpath is not running" > /dev/stderr
			exit 1
		fi
		if [ "$1" != "unlock" ]
		then
			lock_file "$state_lock" 10
			[ "$?" -ne 0 ] && exit 1
			slock="$state_lock"
		fi
	fi

	load_config nh

	ret=0

	while [ "$1" != "" ]
	do
		case "$1" in
		nh_keeparp)
			do_nh_keeparp
			shift 1
			;;
		nh_check)
			do_nh_check
			shift 1
			;;
		nh_down)
			if [ "$2" = "" ]
			then
				echo "Missing NH ID for $1" > /dev/stderr
				ret=1
				break
			fi
			fn="$mpath_data/nh_down.$2"
			if ! [ -e "$fn" ]
			then
				touch "$fn"
				do_nh_changed "$2"
			fi
			shift 1
			shift 1
			;;
		nh_up)
			if [ "$2" = "" ]
			then
				echo "Missing NH ID for $1" > /dev/stderr
				ret=1
				break
			fi
			fn="$mpath_data/nh_down.$2"
			if [ -e "$fn" ]
			then
				rm -f "$fn"
				do_nh_changed "$2"
			fi
			shift 1
			shift 1
			;;
		nh_changed)
			do_nh_changed "$2"
			shift 1
			shift 1
			;;
		nh_list)
			n="${#NHDEV[@]}"
			nln=""
			[ "$n" -le 0 ] && nln=" N/A"
			echo "Next Hops:$nln"
			for ((i=0; i<n; i+=1))
			do
				state="(UP) "
				[ -e "$mpath_data/nh_down.$i" ] && state="(DOWN)"
				gw="${NHGW[$i]}"
				[ "$gw" ] && gw="GW $gw "
				keep="${NHGWKEEP[$i]}"
				[ "$keep" ] && keep=" keep-arp"
				check="${NHGWPING[$i]}"
				[ "$check" ] && check=" ping-check $check"
				arr_nh[$i]="${gw}dev ${NHDEV[$i]}$keep$check"
				echo " - $state ${arr_nh[$i]}"
			done
			shift 1
			;;
		start)
			lock_file "$run_lock" 0
			if [ "$?" -ne 0 ]
			then
				echo "The mpath settings are already applied." > /dev/stderr
				ret=1
				break
			fi
			lock_file "$state_lock" 10
			if [ "$?" -ne 0 ]
			then
				rm -f "$run_lock"
				ret=1
				break
			fi
			slock="$state_lock"
			do_config "add"
			shift 1
			;;
		stop)
			do_config "del"
			rm -f "$state_lock"
			slock=""
			rm -f "$run_lock"
			shift 1
			if [ "$1" != "" ]
			then
				echo "$1: ignored" > /dev/stderr
				ret=1
			fi
			break
			;;
		list)
			list_settings
			shift 1
			;;
		lock)
			slock=""
			shift 1
			if [ "$1" != "" ]
			then
				echo "$1: ignored" > /dev/stderr
			fi
			break
			;;
		unlock)
			slock="$state_lock"
			shift 1
			if [ "$1" != "" ]
			then
				echo "$1: ignored" > /dev/stderr
			fi
			break
			;;
		*)
			echo "Unknown command: $1" > /dev/stderr
			ret=1
			break
			;;
		esac
	done

	[ "$slock" != "" ] && rm -f "$slock"
	exit $ret

else
	# The default action is to show the configuration

	load_config nh

	list_settings

	[ "$1" = "list" ] && exit 0

	# And finally print the supported commands

	echo
	echo "$0 - Manage connections to multiple ISPs"
	echo
	echo "Usage: $0 [command] ..."
	echo
	echo "Commands:"
	echo " list                    - default command, list all settings"
	echo " start                   - configure all defined settings"
	echo " stop                    - unconfigure all defined settings"
	echo " lock                    - lock the settings"
	echo " unlock                  - unlock the settings"
	echo " nh_check                - do the NH state checks and show the changed NHs"
	echo " nh_keeparp              - maintain valid NH state in the ARP cache"
	echo " nh_down <id>            - mark Next Hop DOWN and reconfig"
	echo " nh_up <id>              - mark Next Hop UP and reconfig"
	echo " nh_changed [<id_list>]  - reconfig after Next Hop state change"
	echo " nh_list                 - List all configured Next Hops and their state"
	exit 0
fi

